import json
import math
import numpy as np
import os
import pandas as pd
import datetime

import emod_api.demographics.grid_construction as grid
from emod_api.demographics.BaseInputFile import BaseInputFile
from emod_api.demographics.Node import Node
from emod_api.demographics import DemographicsTemplates as DT
from emod_api.demographics.DemographicsInputDataParsers import node_ID_from_lat_long, duplicate_nodeID_check

# Just make once-static methods module-level functions


def fromBasicNode(lat=0, lon=0, pop=1e6, name=1, forced_id=1):
    """
    This function creates a single-node Demographics instance from the params you give it. 
    """
    new_nodes = [Node(lat=lat, lon=lon, pop=pop, name=name, forced_id=forced_id)]
    return Demographics(nodes=new_nodes)

# MOVE TO demographics/DemographicsInputDataParsers.py
def from_file(base_file):
    with open(base_file, "rb") as src:
        raw = json.load(src)
        nodes = []

        # Load the nodes
        for node in raw["Nodes"]:
            nodes.append(Node.from_data(node))

        # Load the idref
        idref = raw["Metadata"]["IdReference"]

    # Create the file
    return Demographics(nodes, idref, base_file)


def get_node_ids_from_file(demographics_file):
    d = Demographics(demographics_file)
    return sorted(d.node_ids)


def get_node_pops_from_params(tot_pop, num_nodes, frac_rural):
    # Generate node sizes
    nsizes = np.exp( -np.log(np.random.rand(num_nodes - 1)))
    nsizes = frac_rural * nsizes / np.sum(nsizes)
    nsizes = np.minimum(nsizes, 100 / tot_pop)
    nsizes = frac_rural * nsizes / np.sum(nsizes)
    nsizes = np.insert(nsizes, 0, 1 - frac_rural)
    npops = ((np.round(tot_pop * nsizes, 0)).astype(int)).tolist()
    return npops


def from_synth_pop(tot_pop=1e6, num_nodes=100, frac_rural=0.3, id_ref="from_synth_pop"):
    """
    This function creates an EMOD-compatible Demographics object with the population
    and number of nodes specified, distributing per the additional parameters.
    """ 
    npops = get_node_pops_from_params( tot_pop, num_nodes, frac_rural )

    # Generate node lattice
    ucellb = np.array([[1.0, 0.0], [-0.5, 0.86603]])
    nlocs = np.random.rand(num_nodes, 2)
    nlocs[0, :] = 0.5
    nlocs = np.round(np.matmul(nlocs, ucellb), 4)

    nodes = []
    # Add nodes to demographics
    for idx in range(len(npops)):
        nodes.append(Node(forced_id=idx + 1, pop=npops[idx], lat=nlocs[idx, 1], lon=nlocs[idx, 0]))

    return Demographics(nodes=nodes, idref=id_ref)


def _create_grid_files( point_records_file_in, final_grid_files_dir, site ):
    """
    Author: pselvaraj
    Purpose: Create grid file (as csv) from records file. 
    TBD: Needs some attention but probably need a version of this here so we can start from non grid file for spatial DTK input file creation.
    """
    # create paths first...
    output_filename = f"{site}_grid.csv" 
    if not os.path.exists(final_grid_files_dir):
        os.mkdir(final_grid_files_dir)
    out_path = os.path.join(final_grid_files_dir, output_filename )

    if not os.path.exists( out_path ):
        # Then manip data...
        #logging.info("Reading data...")
        print( f"{out_path} not found so we are going to create it." )
        print( f"Reading {point_records_file_in}." )
        point_records = pd.read_csv(point_records_file_in, encoding="iso-8859-1")
        point_records.rename(columns={'longitude': 'lon', 'latitude': 'lat'}, inplace=True)

        if not 'pop' in point_records.columns:
            point_records['pop'] = [5.5] * len(point_records)

        if 'hh_size' in point_records.columns:
            point_records['pop'] = point_records['hh_size']

        # point_records = point_records[point_records['pop']>0]
        x_min, y_min, x_max, y_max = grid.get_bbox(point_records)
        point_records = point_records[
            (point_records.lon >= x_min) & (point_records.lon <= x_max) & (point_records.lat >= y_min) & (
                    point_records.lat <= y_max)]
        gridd, grid_id_2_cell_id, origin, final = grid.construct(x_min, y_min, x_max, y_max)
        gridd.to_csv(os.path.join(final_grid_files_dir, f"{site}_grid.csv"))

        with open(os.path.join(final_grid_files_dir, f"{site}_grid_id_2_cell_id.json"), "w") as g_f:
            json.dump(grid_id_2_cell_id, g_f, indent=3)

        point_records[['gcid', 'gidx', 'gidy']] = point_records.apply(
                grid.point_2_grid_cell_id_lookup,
                args=(grid_id_2_cell_id, origin,), axis=1).apply(pd.Series)

        grid_pop = point_records.groupby(['gcid', 'gidx', 'gidy'])['pop'].apply(np.sum).reset_index()
        grid_pop['pop'] = grid_pop['pop'].apply(lambda x: round(x/5))
        grid_final = pd.merge(gridd, grid_pop, on='gcid')
        grid_final['node_label'] = list(grid_final.index)
        grid_final = grid_final[grid_final['pop'] > 5]
        grid_final.to_csv(os.path.join(final_grid_files_dir, output_filename ))

    print( f"{out_path} gridded population file created or found." )
    return out_path 

def from_csv(input_file, res=30/3600):
    """
    This function uses a csv population-by-node file to create an EMOD-compatible Demographics
    object.
    """
    if not os.path.exists(input_file):
        print(f"{input_file} not found.")
        return

    print( f"{input_file} found and being read for demographics.json file creation." )
    node_info = pd.read_csv(input_file, encoding='iso-8859-1')
    out_nodes = []
    for index, row in node_info.iterrows():
        pop = 0
        if 'under5_pop' in row:
            pop = int(6*row['under5_pop'])
            if pop<25000:
                continue
        else:
            pop = int(row['pop'])
        lat = row['lat']
        lon = row['lon']
        meta = {} 
        """
        meta = {'dot_name': (row['ADM0_NAME']+':'+row['ADM1_NAME']+':'+row['ADM2_NAME']),
                'GUID': row['GUID'],
                'density': row['under5_pop_weighted_density']}
        """

        node = Node(
                lat=lat,
                lon=lon,
                pop=pop,
                name=int(node_ID_from_lat_long(lat, lon, res)),
                forced_id=int(node_ID_from_lat_long(lat, lon, res)), meta=meta)

        out_nodes.append(node)
    out_nodes = duplicate_nodeID_check(out_nodes)
    for node in out_nodes:
        if node.id == 1639001798:  #Not sure why this node causes issues, just dump it for speed.  Probably an issue in the duplicate nodeID check
            remel = node
            out_nodes.remove(remel)

    return Demographics(nodes=out_nodes, idref="from_csv")

def from_pop_csv( pop_filename_in, pop_filename_out="spatial_gridded_pop_dir", site="No_Site" ):
    # This should be a 'raw' ungridded file that needs to be converted into a pop_grid using _create_grid_files from household lat-lons
    # Don't do this if the grid file already exists.
    grid_file_path = _create_grid_files( pop_filename_in, pop_filename_out, site )
    print( f"{grid_file_path} grid file created." )
    return from_csv( grid_file_path )

class Demographics(BaseInputFile):
    """
    This class is a container of data necessary to produce a EMOD-valid demographics input file. It can be initialized from 
    an existing valid demographics.joson type file or from an array of valid Nodes.
    """
    def __init__(self, nodes, idref="Gridded world grump2.5arcmin", base_file=None):
        super(Demographics, self).__init__(idref) 
        self.nodes = nodes
        self.idref = idref
        self.raw = None
        self.implicits = list()

        if base_file:
            with open(base_file, "rb") as src:
                self.raw = json.load(src)
        else:
            meta = self.generate_headers()
            self.raw = {"Metadata": meta, "Defaults": {}}

            self.SetMinimalNodeAttributes()

            self.raw["Defaults"]["IndividualAttributes"] = {}
            # Uniform prevalence between .1 and .3
            #self.raw["Defaults"]["IndividualAttributes"].update( DT.InitPrevUniform() )
            #self.raw["Defaults"]["IndividualAttributes"].update( DT.InitSusceptConstant() )
            self.raw["Defaults"]["IndividualAttributes"].update( DT.NoRisk() )
            #self.raw["Defaults"]["IndividualAttributes"].update( DT.InitAgeUniform() )
            DT.NoInitialPrevalence(self) # does this need to be called?
            #DT.InitSusceptConstant(self) # move this to Measles and maybe other subclasses
            #DT.NoRisk()
            DT.InitAgeUniform(self)

    def generate_file(self, name="demographics.json"):
        self.raw["Nodes"] = []

        #for node in self.nodes.values():
        for node in self.nodes:
            d = {"NodeID": node.id, "NodeAttributes": node.to_dict()}
            d.update(node.meta)
            self.raw["Nodes"].append(d)

        # Update node count
        self.raw["Metadata"]["NodeCount"] = len(self.nodes)
        with open(name, "w") as output:
            json.dump(self.raw, output, indent=3)

        return name

    @property
    def node_ids(self):
        return [node["NodeID"] for node in self.nodes]

    @property
    def node_count(self):
        return len(self.nodes)

    def get_node(self, nodeid):
        """
        Return the node idendified by nodeid. Search either name or actual id
        :param nodeid: 
        :return: 
        """
        if nodeid in self.nodes:
            return self.nodes[nodeid]
        for node in self.nodes.values():
            if node.name == nodeid:
                return node
        raise ValueError(
            "No nodes available with the id: %s. Available nodes (%s)"
            % (nodeid, ", ".join(self.nodes.keys()))
        )

    def SetIndividualAttributtsWithFertMort(self, CrudeBirthRate=40/1000, CrudeMortRate = 20/1000):
        self.raw['Defaults']['IndividualAttributes'] = {}
        self.raw['Defaults']['IndividualAttributes'].update(DT.NoInitialPrevalence())
        self.raw['Defaults']['IndividualAttributes'].update(DT.NoRiskHeterogeneity())

        #Alternative to EveryoneInitiallySusceptible is SimpleSusceptibilityDistribution(meanAgeAtInfection=2.5 or some other number)
        self.raw['Defaults']['IndividualAttributes'].update(DT.EveryoneInitiallySusceptible())
        self.raw['Defaults']['IndividualAttributes'].update(DT.NoMigrationHeterogeneity())
        self.raw['Defaults']['IndividualAttributes'].update(DT.ConstantMortality(CrudeMortRate))
        self.raw['Defaults']['IndividualAttributes'].update(DT.EquilibriumAgeDistFromBirthAndMortRates(CrudeBirthRate, CrudeMortRate))

    def AddIndividualPropertyAndHINT(self, Property, Values, InitialDistribution=None, TransmissionMatrix=None):
        """
        This function takes the HINT configuration json, which are presumed to be correct, and 
        inserts them into the right place in the demoagrphics json struture.
        """
        if 'IndividualProperties' not in self.raw['Defaults']:
            self.raw['Defaults']['IndividualProperties'] = []
        if any([Property in x for x in self.raw['Defaults']['IndividualProperties']]):
            raise ValueError("Property Type '{0}' already present in IndividualProperties list".format(Property))
        else:
            # Check if Property is in whitelist. If not, auto-set Disable_IP_Whitelist
            iplist = ["Age_Bin", "Accessibility", "Geographic", "Place", "Risk", "QualityOfCare", "HasActiveTB", "InterventionStatus"]
            if Property not in iplist:
                print("Need to set Disable_IP_Whitelist in config.")
                def update_config( config ):
                    config.parameters.Disable_IP_Whitelist = 1
                    return config
                if self.implicits is not None:
                    self.implicits.append( update_config )

            if Property == "Age_Bin":
                self.raw['Defaults']['IndividualProperties'].append(
                    {"Property": Property,
                     "Age_Bin_Edges_In_Years": Values,
                     "Transitions": [],
                     "TransmissionMatrix": { "Route": "Contact",
                                             "Matrix": TransmissionMatrix
                    }}) 
            else:
                if TransmissionMatrix is not None:
                    self.raw['Defaults']['IndividualProperties'].append(
                        {"Property": Property,
                         "Values": Values,
                         "Initial_Distribution": InitialDistribution,
                         "TransmissionMatrix": { "Route": "Contact",
                                                 "Matrix": TransmissionMatrix
                        }}) 
                else:
                    self.raw['Defaults']['IndividualProperties'].append(
                        {"Property": Property,
                         "Values": Values,
                         "Initial_Distribution": InitialDistribution
                        }) 
        if TransmissionMatrix is not None:
            def update_config(config):
                config.parameters.Enable_Heterogeneous_Intranode_Transmission = 1
                return config

            if self.implicits is not None:
                self.implicits.append(update_config)

    def AddAgeDependentTransmission(
            self,
            Age_Bin_Edges_In_Years = [0, 1, 2, -1],
            TransmissionMatrix = [[1.0, 1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 1.0, 1.0]]):

        # TBD: Let's call AddIndividualPropertyAndHINT
        self.AddIndividualPropertyAndHINT("Age_Bin", Age_Bin_Edges_In_Years, None, TransmissionMatrix)
        def update_config(config):
            config.parameters.Enable_Heterogeneous_Intranode_Transmission = 1
            return config

        if self.implicits is not None:
            self.implicits.append(update_config)

    def SetDefaultIndividualAttributes(self): 
        """
        NOTE: This is very Measles-ish. We might want to move into MeaslesDemographics
        """
        self.raw['Defaults']['IndividualAttributes'] = {}
        DT.NoInitialPrevalence(self)
        DT.InitRiskUniform( self, 0.5, 1.5 )
        self.raw['Defaults']['IndividualAttributes'].update(DT.NoMigrationHeterogeneity())
        #Age distribution from UNWPP
        DT.AgeStructureUNWPP(self)
        #Mortality rates carried over from Nigeria DHS
        self.raw['Defaults']['IndividualAttributes'].update(DT.MortalityStructureNigeriaDHS())
        DT.DefaultSusceptibilityDistribution(self)

    def SetMinimalNodeAttributes(self): 
        self.SetDefaultNodeAttributes(birth=False)

    def SetDefaultNodeAttributes(self, birth=True):
        self.raw['Defaults']['NodeAttributes'] = {
                    "Altitude": 0,
                    "Airport": 1, # why are these still needed?
                    "Region": 1,
                    "Seaport": 1
                    }
        if birth:
            self.raw['Defaults']['NodeAttributes'].update({
                    "BirthRate": math.log(1.03567)/365
                })

    def SetDefaultIndividualProperties(self):
        self.raw['Defaults']['IndividualProperties'] = []

    def SetDefaultProperties(self): 
        self.SetDefaultNodeAttributes()
        self.SetDefaultIndividualAttributes() #Distributions for initialization of immunity, risk heterogeneity, etc.
        self.SetDefaultIndividualProperties() #Individual properties like accessibility, for targeting interventions

    def SetDefaultPropertiesFertMort(self, CrudeBirthRate = 40/1000, CrudeMortRate = 20/1000):
        """Set defaults assuming constant birth and mortality rate"""

        self.SetDefaultNodeAttributes(CrudeBirthRate)
        self.SetDefaultIndividualAttributes(CrudeBirthRate, CrudeMortRate) #Distributions for initialization of immunity, risk heterogeneity, etc.
        self.SetDefaultIndividualProperties() #Individual properties like accessibility, for targeting interventions

    def SetDefaultFromTemplate(self, template, setter_fn=None):
        # TBD: Add some error checking. Make sure IndividualAttributes are Individual, not individual.
        self.raw['Defaults']['IndividualAttributes'].update(template)
        if self.implicits is not None and setter_fn is not None:
            self.implicits.append(setter_fn)

    def SetNodeDefaultFromTemplate(self, template, setter_fn):
        # TBD: Add some error checking. Make sure NodeAttributes are Node, not individual.
        self.raw['Defaults']['NodeAttributes'].update(template)
        if self.implicits is not None:
            self.implicits.append(setter_fn)

    def SetOverdispersion( self, new_overdispersion_value, nodes=[] ):
        def enable_overdispersion( config ):
            print( "DEBUG: Setting 'Enable_Infection_Rate_Overdispersion' to 1." )
            config.parameters.Enable_Infection_Rate_Overdispersion = 1
            return config
        if self.implicits is not None:
            self.implicits.append( enable_overdispersion )
        self.raw['Defaults']['NodeAttributes']["InfectivityOverdispersion"] = new_overdispersion_value

    def SetConstantSusceptibility( self ):
        DT.InitSusceptConstant(self)
