import numpy as np
import math
import scipy.sparse        as sp
import scipy.sparse.linalg as la
import pandas as pd
from copy import deepcopy

def set_suscept_complex( config ):
    config.parameters.Susceptibility_Initialization_Distribution_Type = "DISTRIBUTION_COMPLEX"
    return config

def set_suscept_simple( config ):
    config.parameters.Susceptibility_Initialization_Distribution_Type = "DISTRIBUTION_SIMPLE"
    return config

def set_age_simple( config ):
    config.parameters.Age_Initialization_Distribution_Type = "DISTRIBUTION_SIMPLE"
    return config

def set_age_complex( config ):
    config.parameters.Age_Initialization_Distribution_Type = "DISTRIBUTION_COMPLEX"
    return config

def set_init_prev( config ):
    config.parameters.Enable_Initial_Prevalence = 1
    return config

def set_boxcar_seasonal( config ):
    #config.parameters.Enable_Infectivity_Scaling = 1 # implicit
    config.parameters.Enable_Infectivity_Scaling_Boxcar = 1
    return config

# 
# Risk
#
"""
This submodule contains a bunch of json blobs that are valid pieces of a demographics json file.
They serve as presets or templates to help build up a demographics.json without having to know 
the json. This doesn't represent a full solution from dev team pov but the encapsulation and 
abstraction of this API does at least provide a good step in the right direction.
"""
def NoRisk(): 
    """
    NoRisk puts everyone at 0 risk.
    """
    return {"RiskDist_Description": "No risk",
            "RiskDistributionFlag": 0, # 0 = CONSTANT
            "RiskDistribution1": 0,
            "RiskDistribution2": 0}

def FullRisk( demog ): 
    """
    FullRisk puts everyone at 100% risk.
    """
    setting = {"RiskDist_Description": "Full risk",
            "RiskDistributionFlag": 0, # 0 = CONSTANT
            "RiskDistribution1": 1,
            "RiskDistribution2": 0}
    demog.SetDefaultFromTemplate( setting )

def InitRiskUniform( demog, min=0, max=1 ):
    """
    InitRiskUniform puts everyone at somewhere between 0% risk and 100% risk, drawn uniformly.

    Args:
        min (float): Low end of uniform distribution. Must be >=0, <1.
        max (float): High end of uniform distribution. Must be >=min, <=1.

    Returns:
        json object aka python dict that can be directly passed to Demographics::SetDefaultFromTemplate

    Raises:
        None

    """
    setting = {"RiskDistributionFlag": 1,
            "RiskDistribution1": min,
            "RiskDistribution2": max }
    demog.SetDefaultFromTemplate( setting )
# 
# Initial Prevalence
# config: Enable_Initial_Prevalence=1
#
def NoInitialPrevalence( demog ):
    """
    NoInitialPrevalence disables initial prevalence; outbreak seeding must be done from an Outbreak intervention (or serialized population).

    Args:
        demog: emod-api.demographics.Demographics instance.

    Returns:
        None

    Raises:
        None
    """

    setting = {"PrevalenceDist_Description": "No initial prevalence",
               "InitialPrevalence": 0,
              }
    # why not just disable it at config?
    demog.SetDefaultFromTemplate( setting )
# 
# Initial Susceptibility (1-Immunity)
#

def InitPrevUniform( demog, prevalence ):
    # old
    setting = {"PrevalenceDist_Description": f"{prevalence} initial prevalence",
               "PrevalenceDistributionFlag": 0,
               "PrevalenceDistribution1": prevalence,
               "PrevalenceDistribution2": 0 }
    # new
    setting.update( { "InitialPrevalence": prevalence } )
    demog.SetDefaultFromTemplate( setting, set_init_prev )

def InitSusceptConstant( demog ):
    # set config.Susceptibility_... SIMPLE
    setting = {"SusceptibilityDistributionFlag": 0,
            "SusceptibilityDistribution1": 1,
            "SusceptibilityDistribution2": 0 }
    demog.SetDefaultFromTemplate( setting, set_suscept_simple )

def EveryoneInitiallySusceptible( demog, setting=1.0 ):
    # set config.Susceptibility_Initialization_Distribution_Type=COMPLEX
    suscDist = {
        "SusceptibilityDist_Description": f"Everyone is initially susceptible with probability {setting}",
        "SusceptibilityDistribution": {
            "DistributionValues": [
                [0, 36500]
            ],
            "ResultScaleFactor": 1,
            "ResultValues": [
                [setting, setting]
            ]
        }
    }
    demog.SetDefaultFromTemplate( suscDist, set_suscept_complex )

def StepFunctionSusceptibility( demog, protected_setting=0.0, threshold_age=365*5.0 ):
    # set config.Susceptibility_Initialization_Distribution_Type=COMPLEX
    suscDist = {
        "SusceptibilityDist_Description": "Youngers are somewhat protected",
        "SusceptibilityDistribution": {
            "DistributionValues": [
                [0, threshold_age, threshold_age, 36500]
            ],
            "ResultScaleFactor": 1,
            "ResultValues": [
                [protected_setting, protected_setting, 1.0, 1.0]
            ]
        }
    }
    demog.SetDefaultFromTemplate( suscDist, set_suscept_complex )

def SimpleSusceptibilityDistribution( demog, meanAgeAtInfection=2.5): 
    # set config.Susceptibility_Initialization_Distribution_Type=COMPLEX
    # This function is first to be switched over to be reversed.
    # Calling code in emodpy will call this and pass the demographics instance then we
    # call SetDefaultFromTemplate on the demog object so we can also pass the setter function
    suscDist = {
        "SusceptibilityDist_Description": f"Rough initialization to reduce burn-in and prevent huge outbreaks at sim start.  Exponential distribution, Average age at infection ~{meanAgeAtInfection} years, minimum susceptibility is 2.5% at old ages",
        "SusceptibilityDistribution": {
            "DistributionValues": [
                [i * 365 for i in range(100)]
            ],
            "ResultScaleFactor": 1,
            "ResultValues": [
                [1.0, 1.0] + [0.025 + 0.975 * math.exp(-(i - 1) / (meanAgeAtInfection / math.log(2))) for i in range(2, 100, 1)]
            ]
        }
    }
    demog.SetDefaultFromTemplate( suscDist, set_suscept_complex )

def DefaultSusceptibilityDistribution( demog ): 
    # set config.Susceptibility_Initialization_Distribution_Type=COMPLEX
    suscDist = {
        "SusceptibilityDist_Description": "Rough initialization to reduce burn-in and prevent huge outbreaks at sim start.  Exponential distribution, Average age at infection ~3.5 years, minimum susceptibility is 2.5% at old ages",
        "SusceptibilityDistribution": {
            "DistributionValues": [
                [i * 365 for i in range(100)]
            ],
            "ResultScaleFactor": 1,
            "ResultValues": [
                [1.0, 1.0] + [0.025 + 0.975 * math.exp(-(i - 1) / (2.5 / math.log(2))) for i in range(2, 100, 1)]
            ]
        }
    }
    demog.SetDefaultFromTemplate( suscDist, set_suscept_complex )

#
# Mortality
#
def MortalityRateByAge(AgeBins, MortRate):
    """Mortality Rate by Age"""
    return {"MortalityDistribution": {
        "NumDistributionAxes": 2,
        "AxisNames": ["gender", "age"],
        "AxisUnits": ["male=0,female=1", "years"],
        "AxisScaleFactors": [1, 365],
        "NumPopulationGroups": [2, 1],
        "PopulationGroups": [
            [0, 1],
            AgeBins
        ],
        "ResultScaleFactor": 1,
        "ResultUnits": "daily probability of dying",
        "ResultValues": [
            MortRate,
            MortRate
        ]
    }
    }

def ConstantMortality(CrudeMortRate):
    MortRate = -1 * (math.log(1 - CrudeMortRate) / 365)
    return {"MortalityDistribution": {
        "NumDistributionAxes": 2,
        "AxisNames": ["gender", "age"],
        "AxisUnits": ["male=0,female=1", "years"],
        "AxisScaleFactors": [1, 365],
        "NumPopulationGroups": [2, 1],
        "PopulationGroups": [
            [0, 1],
            [0]
        ],
        "ResultScaleFactor": 1,
        "ResultUnits": "daily probability of dying",
        "ResultValues": [
            [MortRate],
            [MortRate]
        ]
    }
    }

def MortalityStructureNigeriaDHS():
    morts = [ 0,0.0019158015385118965,0.0019158015385118965,0.0001717560763629944, 0.0001717560763629944, 4.9704718676046866e-05,4.9704718676046866e-05,5.534972988163744e-06,5.534972988163744e-06, 1.0006476080515192e-05,1.0006476080515192e-05,0.0003153728749953899,0.0003153728749953899,0.99]
    return {
        "MortalityDistribution": {
        "NumDistributionAxes": 2,
        "AxisNames": ["gender","age"],
        "AxisUnits": ["male=0,female=1","years"],
        "AxisScaleFactors": [1,365],
        "NumPopulationGroups": [2,14],
        "PopulationGroups": [
            [0,1],
            [0,0.0001,0.08082191780821918,0.08092191780821918,1,1.0001,5,5.0001,15,15.0001,50,50.0001,90,90.0001]
        ],
        "ResultScaleFactor":1,
        "ResultUnits": "daily probability of dying",
        "ResultValues": [morts, morts]
        }
    }

#
# Age Structure
#
def InitAgeUniform( demog ):
    setting = { "AgeDistributionFlag": 1,
                "AgeDistribution1": 0,
                "AgeDistribution2": 18250 }
    demog.SetDefaultFromTemplate( setting, set_age_simple )

def _computeAgeDist(bval,mvecX,mvecY,fVec):
    """compute equilibrium age distribution given age-specific mortality and crude birth rates

    :param bval: crude birth rate in births per day per person
    :param mvecX: list of age bins in days
    :param mvecY: List of per day mortality rate for the age bins
    :param fVec: Seasonal forcing per month


    returns ??, MonthlyAgeDist, MonthlyAgeBins
    author: Kurt Frey
    """

    max_yr = 90
    bin_size = 30
    day_to_year = 365

    # Age brackets
    avecY = np.arange(0, max_yr * day_to_year, bin_size) - 1

    # Mortality sampling
    mvecX = [-1] + mvecX + [max_yr * day_to_year + 1]
    mvecY = [mvecY[0]] + mvecY + [mvecY[-1]]
    mX = np.arange(0, max_yr * day_to_year, bin_size)
    mX[0] = 1
    mval = 1.0 - np.interp(mX, xp=mvecX, fp=mvecY)
    r_n = mval.size

    # Matrix construction
    BmatRC = (np.zeros(r_n), np.arange(r_n))
    Bmat = sp.csr_matrix(([bval * bin_size] * r_n, BmatRC), shape=(r_n, r_n))
    Mmat = sp.spdiags(mval[:-1] ** bin_size, -1, r_n, r_n)
    Dmat = Bmat + Mmat

    # Math
    (gR, popVec) = la.eigs(Dmat, k=1, sigma=1.0)
    gR = np.abs(gR ** (float(day_to_year) / float(bin_size)))
    popVec = np.abs(popVec) / np.sum(np.abs(popVec))

    # Apply seasonal forcing
    mVecR = [-2.0, 30.5, 30.6, 60.5, 60.6, 91.5, 91.6, 121.5,
             121.6, 152.5, 152.6, 183.5, 183.6, 213.5, 213.6, 244.5,
             245.6, 274.5, 274.6, 305.5, 305.6, 333.5, 335.6, 364.5]
    fVec = np.flipud([val for val in fVec for _ in (0, 1)])
    wfVec = np.array([np.mean(np.interp(np.mod(range(val + 1, val + 31), 365),
                                        xp=mVecR, fp=fVec)) for val in avecY]).reshape(-1, 1)
    popVec = popVec * wfVec / np.sum(popVec * wfVec)

    # Age sampling
    avecY[0] = 0
    avecX = np.clip(np.around(np.cumsum(popVec), decimals=7), 0.0, 1.0)
    avecX = np.insert(avecX, 0, np.zeros(1))

    return gR.tolist()[0], avecX[:-1].tolist(), avecY.tolist()

def AgeStructureUNWPP( demog ):
    setting = {
            "AgeDistribution": {
                "NumDistributionAxes": 0,
                "ResultUnits": "years",
                "ResultScaleFactor": 365,
                "ResultValues": [0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90],
                "DistributionValues": [0, 0.164, 0.3077, 0.4317, 0.5381, 0.6289, 0.7066, 0.7721, 0.8256, 0.8690, 0.9036,
                                       0.9316, 0.9537, 0.9708, 0.9833, 0.9918, 0.9968, 0.9991, 1.0]
                }
            }
    demog.SetDefaultFromTemplate( setting, set_age_complex )

def EquilibriumAgeDistFromBirthAndMortRates( demog, CrudeBirthRate=40/1000, CrudeMortRate=20/1000 ):
    # For the moment, this is assuming constant mortality, should update with an interpolator based on demographics file to improve
    # Reuse of some of these demographics variables and passing them around indicates this may be better as a class.  May be
    # best to wait however, until the idmtools Demographics class is around, so that we can inherit from that and only add functionality.
    # Should maybe allow user to specify the "fineness" of the age-binning for the final demographics distribution
    # Having to do this conversion of crude birth rates to per-day per-person rates frequently.  Should find the right place
    # to do this only once.
    BirthRate = math.log(1 + CrudeBirthRate) / 365
    MortRate = -1 * (math.log(1 - CrudeMortRate) / 365)

    # It is important for the age distribution computation that the age-spacing be very fine; I've used 30 days here.
    # With coarse spacing, the computation in practice doesn't work as well.
    ageDist = _computeAgeDist(BirthRate, [i * 30 for i in range(1200)], 1200 * [MortRate], 12 * [1.0])

    # The final demographics file, though, can use coarser binning interpolated from the finely-spaced computed distribution.
    EMODAgeBins = list(range(16)) + [20+5*i for i in range(14)]
    EMODAgeDist = np.interp(EMODAgeBins, [i/365 for i in ageDist[2]], ageDist[1]).tolist()
    EMODAgeBins.extend([90])
    EMODAgeDist.extend([1.0])
    setting = {"AgeDistribution": {
            "NumDistributionAxes": 0,
            "ResultUnits": "years",
            "ResultScaleFactor": 365,
            "ResultValues": EMODAgeBins,
            "DistributionValues": EMODAgeDist
            }
        }
    demog.SetDefaultFromTemplate( setting, set_age_complex )

def AddSeasonalForcing( demog, start=100, end=330, factor=1.0 ):
    """enable seasonal forcing on the default node attributes via boxcar scaling

    Requires config enables infectivity scaling: e.g.:
    'Enable_Infectivity_Scaling': 1,
    'Enable_Infectivity_Scaling_Boxcar': 1,
    """

    setting = {
        "InfectivityMultiplier": 1.0,
        "InfectivityBoxcarAmplitude": factor,
        "InfectivityBoxcarStartTime": start,
        "InfectivityBoxcarEndTime": end}
    demog.SetNodeDefaultFromTemplate( setting, set_boxcar_seasonal )


    
# def MinimalNodeAttributes():
#    TBD

#
# MigrationHeterogeneity -- Deleted in G-O
#
def NoMigrationHeterogeneity():
    return {"MigrationDist_Description": "No migration heterogeneity",
            "MigrationHeterogeneityDistributionFlag": 0,
            "MigrationHeterogeneityDistribution1": 1,
            "MigrationHeterogeneityDistribution2": 0}

