import emod_api.interventions.utils as utils
from emod_api import schema_to_class as s2c

import copy

schema_path=None

###
### Generic
###
def BroadcastEvent(
        camp,
        Event_Trigger: str = 'Births'
    ):
    """
        Wrapper function to create and return a BroadcastEvent intervention.

        Args:
            camp: emod_api.campaign object with schema_path set.
            Event_Trigger: A valid trigger/event/signal.

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            BroadastEvent intervention ready to be added to a campaign.
    """
    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )
    intervention = s2c.get_class_with_defaults( "BroadcastEvent", schema_path )
    intervention.Broadcast_Event = camp.get_event( Event_Trigger, old=True )
    return intervention


def BroadcastEventToOtherNodes(
        camp,
        Event_Trigger,
        Node_Selection_Type="DISTANCE_ONLY",
        Max_Distance_To_Other_Nodes_Km=-1,
        Include_My_Node=1
    ):
    """
        Wrapper function to create and return a BroadcastEventToOtherNodes intervention. 

        Args:
            camp: emod_api.campaign object with schema_path set.
            Event_Trigger: A valid trigger/event/signal.
            Node_Selection_Type: TBD.
            Max_Distance_To_Other_Nodes_Km: TBD.
            Include_My_Node: TBD.

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            BroadastEvent intervention ready to be added to a campaign.
    """
    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )
    intervention = s2c.get_class_with_defaults( "BroadcastEventToOtherNodes", schema_path )
    intervention.Event_Trigger = camp.get_event( Event_Trigger, old=True )
    intervention.Node_Selection_Type = Node_Selection_Type 
    if Max_Distance_To_Other_Nodes_Km != -1:
        intervention.Max_Distance_To_Other_Nodes_Km = Max_Distance_To_Other_Nodes_Km 
    intervention.Include_My_Node = Include_My_Node 
    return intervention

def MultiInterventionDistributor(
        camp,
        Intervention_List
    ):
    """
        Wrapper function to create and return a MultiInterventionDistributor intervention. 

        Args:
            camp: emod_api.campaign object with schema_path set.
            Intervention_List: List of 1 or more valid intervention dictionaries to be
            distributed together. 

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            MultiInterventionDistributor intervention ready to be added to a campaign.
    """
   
    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )
    if Intervention_List is None or type(Intervention_List) is not list or len( Intervention_List ) == 0:
        raise ValueError( "Intervention_List is empty or None or not a list." )
    intervention = s2c.get_class_with_defaults( "MultiInterventionDistributor", schema_path )
    intervention.Intervention_List = copy.deepcopy( Intervention_List )
    return intervention


def DelayedIntervention(
        camp,
        Configs,
        Delay_Dict=None
    ):
    """
        Wrapper function to create and return a DelayedIntervention intervention. 

        Args:
            camp: emod_api.campaign object with schema_path set.  
            Config: Valid intervention config.
            Delay_Dict: Dictionary of 1 or 2 params that are the literal Delay_Distribution
            parameters, but without the distribution, which is inferred. E.g., 
            { "Delay_Period_Exponential": 5 }

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            DelayedIntervention intervention ready to be added to a campaign.
    """
    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )

    if Configs is None or type( Configs ) is not list or len( Configs ) == 0:
        raise ValueError( "Configs is empty or None or not a list." )

    intervention = s2c.get_class_with_defaults( "DelayedIntervention", schema_path )
    intervention.Actual_IndividualIntervention_Configs = copy.deepcopy( Configs )
    if Delay_Dict is None:
        Delay_Dict = { "Delay_Period_Constant": 0 }
    for param in Delay_Dict:
        setattr( intervention, param, Delay_Dict[ param ] )
    return intervention


def HSB(
        camp,
        Event_Or_Config="Event",
        Config=None,
        Event="NoTrigger",
        Tendency=1.0,
        Single_Use=True,
        Name="HSB"
    ):
    """
        Wrapper function to create and return a HealthSeekingBehaviour intervention. 

        Args:
            camp: emod_api.campaign object with schema_path set.
            Event_Or_Config: "Event" or "Config".
            Config: Complete, valid intervention configuration to be distributed.
            Event: Event/Trigger/Signal to be broadcast, alternative to an intervention.
            Tendency: Daily probability of 'seeking care' aka distributing payload intervention.
            Single_Use: One-and-done, or continuous?
            Name: Intervention Name. Useful if you want to provide uniqueness and not worry about 
            duplicate intervention management.

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            HSB intervention ready to be added to a campaign.
    """
    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )

    intervention = s2c.get_class_with_defaults( "SimpleHealthSeekingBehavior", schema_path )
    intervention.Event_Or_Config = Event_Or_Config
    if intervention.Event_Or_Config == "Event":
        intervention.Actual_IndividualIntervention_Event = camp.get_event( Event, old=True )
    else:
        if Config is None:
            raise ValueError( "You specified 'Config' but no actual Config." )
        intervention.Actual_IndividualIntervention_Config = Config
    intervention.Tendency = Tendency 
    intervention.Intervention_Name = Name 
    intervention.Single_Use = Single_Use 
    return intervention


def NLHTI(
        camp,
        Triggers,
        Interventions,
        Property_Restrictions=None,
        Demographic_Coverage=1.0,
        Target_Demographic="Everyone",
        Target_Age_Min=0,
        Target_Age_Max=125*365,
        Target_Gender="All",
        Target_Residents_Only=1,
        Duration=-1,
        Blackout_Event_Trigger="Blackout_Event_Trigger",
        Blackout_Period=0,
        Blackout_On_First_Occurrence=0,
        Disqualifying_Properties=None,
    ):
    """
        Wrapper function to create and return a NodeLevelHealthTriggeredIntervention intervention. 

        Args:
            camp: emod_api.campaign object with schema_path set.
            Triggers: List of Triggers/Events/Signals
            Interventions: List of interventions to distrbute when signal is heard.
            Property_Restrictions: Individual Properties that an agent must have to qualify for intervention.
            Demographic_Coverage: Percentage of individuals to receive intervention.
            Target_Demographic: Everyone, ExplicitAgeRanges, etc.
            Target_Age_Min: Minimum age (in years).
            Target_Age_Max: Maximum age (in years).
            Target_Gender: All, Male, or Female.
            Target_Residents_Only: Not used.
            Duration: How long this listen-and-distribute should last.
            Blackout_Event_Trigger: Not used.
            Blackout_Period: Not used.
            Blackout_On_First_Occurrence: Not used.
            Disqualifying_Properties: Not used.

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            NLHTI intervention ready to be added to a campaign.
    """
    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )

    intervention = s2c.get_class_with_defaults( "NodeLevelHealthTriggeredIV", schema_path )
    intervention.Trigger_Condition_List = [ camp.get_event( trigger, old=True ) for trigger in Triggers ]
    #intervention.Node_Property_Restrictions = Node_Property_Restrictions 
    #intervention.Property_Restrictions_Within_Node = Property_Restrictions_Within_Node 
    intervention.Target_Demographic = Target_Demographic 
    intervention.Target_Age_Min = Target_Age_Min 
    intervention.Target_Age_Max = Target_Age_Max
    intervention.Target_Gender = Target_Gender 
    intervention.Demographic_Coverage = Demographic_Coverage 
    intervention.Duration = Duration 
    intervention.Actual_IndividualIntervention_Config = MultiInterventionDistributor( camp, Interventions )

    return intervention

 
def PropertyValueChanger(
        camp,
        Target_Property_Key,
        Target_Property_Value,
        New_Property_Value,
        Daily_Probability=1.0,
        Maximum_Duration=1,
        Revert=False,
        Intervention_Name="",
        Event_Trigger_Distributed="",
        Event_Trigger_Expired=""
    ):
    """
        Wrapper function to create and return a PropertyValueChanger intervention. 

        Args:
            camp: emod_api.campaign object with schema_path set. 

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            PropertyValueChanger intervention ready to be added to a campaign.
    """
    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )

    intervention = s2c.get_class_with_defaults( "PropertyValueChanger", schema_path )
    intervention.Target_Property_Key = Target_Property_Key
    intervention.Target_Property_Value = Target_Property_Value
    intervention.New_Property_Value = New_Property_Value
    intervention.Daily_Probability = Daily_Probability 
    intervention.Maximum_Duration=Maximum_Duration
    intervention.Revert=Revert
    if len(Intervention_Name)>0:
        intervention.Intervention_Name=Intervention_Name
    if len(Event_Trigger_Distributed)>0:
        intervention.Event_Trigger_Distributed=Event_Trigger_Distributed
    if len(Event_Trigger_Expired)>0:
        intervention.Event_Trigger_Expired=Event_Trigger_Expired

    return intervention


def ScheduledCampaignEvent(
        camp,
        Start_Day,
        Nodeset_Config,
        Number_Repetitions=1,
        Timesteps_Between_Repetitions=-1,
        Event_Name="Scheduled_Campaign_Event",
        Node_Property_Restrictions=None,
        Property_Restrictions=None,
        Target_Demographic="Everyone",
        Demographic_Coverage=1.0,
        Target_Age_Min=0,
        Target_Age_Max=125*365,
        Target_Gender="All",
        Intervention_List=None):

    """
        Wrapper function to create and return a ScheduledCampaignEvent intervention. 
        The alternative to a ScheduledCampaignEvent is a TriggeredCampaignEvent.

        Args:
            camp: emod_api.campaign object with schema_path set.
            Start_Day: When to start.
            Event_Name: Name for overall campaign event, of not functional meaning.
            Nodeset_Config: Nodes to target with this intervenion, return from utils.do_nodes().
            Node_Property_Restrictions: Optional property restriction for node-targeted interventions.
            Property_Restrictions: Individual Properties a person must have to receive the intervention(s).
            Number_Repetitions: N/A 
            Timesteps_Between_Repetitions: N/A 
            Demographic_Coverage: Percentage of individuals to receive intervention.
            Target_Demographic: Everyone, ExplicitAgeRanges, etc.
            Target_Age_Min: Minimum age (in years).
            Target_Age_Max: Maximum age (in years).
            Target_Gender: All, Male, or Female.
            Intervention_List: List of 1 or more valid intervention dictionaries to be
            distributed together. 

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            ScheduledCampaignEvent intervention ready to be added to a campaign.
    """
    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )

    # Not checking Intervention_list because MultiInterventionDistributor checks

    if Node_Property_Restrictions is None:
        Node_Property_Restrictions=[]
    if Property_Restrictions is None:
        Property_Restrictions=[]

    event = s2c.get_class_with_defaults( "CampaignEvent", schema_path )
    coordinator = s2c.get_class_with_defaults( "StandardEventCoordinator", schema_path )
    coordinator.Demographic_Coverage = Demographic_Coverage

    # Second, hook them up
    event.Event_Coordinator_Config = coordinator
    event.Start_Day = float(Start_Day)
    event.Nodeset_Config = Nodeset_Config # utils.do_nodes( schema_path, node_ids )

    coordinator.Intervention_Config = MultiInterventionDistributor( camp, Intervention_List )
    #coordinator.Node_Property_Restrictions = Node_Property_Restrictions 
    #coordinator.Property_Restrictions_Within_Node = Property_Restrictions_Within_Node 
    coordinator.Target_Demographic = Target_Demographic 
    coordinator.Demographic_Coverage = Demographic_Coverage 
    coordinator.Target_Age_Min = Target_Age_Min 
    coordinator.Target_Age_Max = Target_Age_Max
    coordinator.Target_Gender = Target_Gender 
    
    return event


def TriggeredCampaignEvent(
        camp,
        Start_Day,
        Event_Name,
        Nodeset_Config,
        Triggers,
        Intervention_List,
        Node_Property_Restrictions=None,
        Property_Restrictions=None,
        Number_Repetitions=1,
        Timesteps_Between_Repetitions=-1,
        Demographic_Coverage=1.0,
        Target_Demographic="Everyone",
        Target_Age_Min=0,
        Target_Age_Max=125*365,
        Target_Gender="All",
        Target_Residents_Only=1,
        Duration=-1,
        Blackout_Event_Trigger="Blackout_Event_Trigger",
        Blackout_Period=0,
        Blackout_On_First_Occurrence=0,
        Disqualifying_Properties=None,
        Delay=None
    ):
    """
        Wrapper function to create and return a TriggeredCampaignEvent intervention. 
        The alternative to a TriggeredCampaignEvent is a ScheduledCampaignEvent.

        Args:
            camp: emod_api.campaign object with schema_path set.
            Start_Day: When to start.
            Event_Name: Name for overall campaign event, of not functional meaning.
            Nodeset_Config: Nodes to target with this intervenion, return from utils.do_nodes().
            Triggers: List of triggers/events/signals to listen to in order to trigger distribution.
            Intervention_List: List of 1 or more valid intervention dictionaries to be
            distributed together. 
            Node_Property_Restrictions: N/A.
            Property_Restrictions: Individual Properties a person must have to receive the intervention(s).
            Number_Repetitions: N/A 
            Timesteps_Between_Repetitions: N/A 
            Demographic_Coverage: Percentage of individuals to receive intervention.
            Target_Demographic: Everyone, ExplicitAgeRanges, etc.
            Target_Age_Min: Minimum age (in years).
            Target_Age_Max: Maximum age (in years).
            Target_Gender: All, Male, or Female.
            Target_Residents_Only: Not used.
            Duration: How long this listen-and-distribute should last.
            Blackout_Event_Trigger: Not used.
            Blackout_Period: Not used.
            Blackout_On_First_Occurrence: Not used.
            Disqualifying_Properties: Not used.
            delay: Optional delay between trigger and actual distribution.

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            TriggeredCampaignEvent intervention ready to be added to a campaign.
    """
    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )

    if Node_Property_Restrictions is None:
        Node_Property_Restrictions=[]
    if Property_Restrictions is None:
        Property_Restrictions=[]

    event = s2c.get_class_with_defaults( "CampaignEvent", schema_path )
    event.Start_Day = float(Start_Day) 
    event.Nodeset_Config = Nodeset_Config # utils.do_nodes( schema_path, node_ids )

    coordinator = s2c.get_class_with_defaults( "StandardEventCoordinator", schema_path )

    if Delay is not None:
        Intervention_List = [ DelayedIntervention( camp, Intervention_List, Delay_Dict = { "Delay_Period_Constant": Delay } ) ]

    intervention = NLHTI(
        camp,
        Triggers,
        Intervention_List,
        Duration=Duration,
        Property_Restrictions = Property_Restrictions,
        #Number_Repetitions = Number_Repetitions,
        #Timesteps_Between_Repetitions = Timesteps_Between_Repetitions,
        Target_Demographic = Target_Demographic,
        Demographic_Coverage = Demographic_Coverage,
        Target_Age_Min = Target_Age_Min,
        Target_Age_Max = Target_Age_Max,
        Target_Gender = Target_Gender 
    )

    event.Event_Coordinator_Config = coordinator
    coordinator.Intervention_Config = intervention
    return event


# M-O only right now...
def StandardDiagnostic(
        camp,
        Base_Sensitivity=1,
        Base_Specificity=1,
        Days_To_Diagnosis=0,
        Treatment_Fraction=1,
        Event_Or_Config="Config"
    ):
    """
        Wrapper function to create and return a StandardDiagnostic intervention. 

        Args:
            camp: emod_api.campaign object with schema_path set.
            Intervention_List: List of 1 or more valid intervention dictionaries to be
            distributed together. 

        Returns:
            ReadOnlyDict: Schema-based smart dictionary representing a new 
            MultiInterventionDistributor intervention ready to be added to a campaign.
    """

    global schema_path
    schema_path = ( camp.schema_path if camp is not None else schema_path )

    # First, get the objects
    intervention = s2c.get_class_with_defaults( "StandardDiagnostic", schema_path )
    intervention.Base_Sensitivity = Base_Sensitivity 
    intervention.Base_Specificity = Base_Specificity 
    intervention.Detection_Threshold = Detection_Threshold 
    intervention.Event_Or_Config = Event_Or_Config

    return intervention


